/*
 * Copyright (c) 2009 Alex Hornung.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>

#include <string.h>

#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/types.h>

#include <sys/stat.h>

#include <libnotify/notify.h>

#include <signal.h>
#include "thinkpad_notifications.h"

#include <sys/inotify.h>

struct fdfiles active_fd[] = {
                              { "/proc/acpi/ibm/volume",                -1, -1 },
                              { "/proc/acpi/ibm/brightness",            -1, -1 },
                              { "/proc/acpi/video/VID/LCD0/brightness", -1, -1 }
                             };
struct fdselect fds;

void build_select_list() {
  FD_ZERO(&fds.triggeredfd);
  FD_SET(fds.inotifyfd, &fds.triggeredfd);
}

void build_watch_list() {
  int rc;
  int listnum;

  for(listnum = 0; listnum < fds.nofd; listnum++)
    rc = inotify_add_watch_fdselect(fds.activefd[listnum].pathname, IN_ACCESS, listnum);
}

void show_notify_volume(gint volume, int mute) {
  char		name[12]        = "\0";

  static gint 	volume_old 	= 255;

  static int 	mute_old 	= -1;

  NotifyNotification *not;

  volume = (gint) ((int) volume)*100/14;
  sprintf(name, "Volume %d%%", (int) volume );

  if (!notify_init(name)) {
    printf("no good at notify_init()\n");
    return;
  }

  if (mute)
    not = notify_notification_new(name, NULL, "notification-audio-volume-muted", NULL);
  else {
    if (volume > 80)
      not = notify_notification_new(name, NULL, "notification-audio-volume-high", NULL);

    else if (volume > 40)
      not = notify_notification_new(name, NULL, "notification-audio-volume-medium", NULL);

    else if (volume > 0)
      not = notify_notification_new(name, NULL, "notification-audio-volume-low", NULL);

    else
      not = notify_notification_new(name, NULL, "notification-audio-volume-off", NULL);
  }

  notify_notification_set_hint_int32(not, "value", volume);
  notify_notification_set_hint_string(not, "x-canonical-private-synchronous", "");

  notify_notification_set_timeout(not, NOTIFICATION_TIMEOUT);
  notify_notification_show(not, NULL);

  g_object_unref(G_OBJECT(not));

  volume_old = volume;

  mute_old = mute;
}

void show_notify_brightness(gint brightness) {
  char		name[16] 	= "\0";

  static gint 	brightness_old 	= 255;

  NotifyNotification *not;

  brightness = (gint) ((int) brightness)*100/7;
  sprintf(name, "Brightness %d%%", (int) brightness);

  if (!notify_init(name)) {
    printf("no good at notify_init()\n");
    return;
  }

  if (brightness > 80)
    not = notify_notification_new(name, NULL, "notification-display-brightness-high", NULL);

  else if (brightness > 40)
    not = notify_notification_new(name, NULL, "notification-display-brightness-medium", NULL);

  else if (brightness > 0)
    not = notify_notification_new(name, NULL, "notification-display-brightness-low", NULL);

  else
    not = notify_notification_new(name, NULL, "notification-display-brightness-off", NULL);

  notify_notification_set_hint_int32(not, "value", brightness);
  notify_notification_set_hint_string(not, "x-canonical-private-synchronous", "");

  notify_notification_set_timeout(not, NOTIFICATION_TIMEOUT);
  notify_notification_show(not, NULL);

  g_object_unref(G_OBJECT(not));

  brightness_old = brightness;

}

int volume_stuff(int fd) {
  int nread,
      mute;

  gint volume;

  char buf[1024];

  char *bufptr,
       *offset;

  if ( (nread = read(fd, buf, 1024)) > 1) {

    lseek(fd, 0, SEEK_SET);
    buf[nread] = 0x0;

    bufptr = buf;
    offset = strchr(bufptr, '\n');

    if (offset != NULL) {
      for (; !isdigit(*bufptr) && (*bufptr != 0x0); bufptr++);

      if (*bufptr == 0x0) {
        printf("volume crap occured\n");
        return -1;
      }
      volume = g_ascii_strtoull(bufptr, NULL, 10);

      for (offset++; (*offset != 'o') && (*offset != 0x0); offset++);

      if (*offset == 0x0) {
        printf("more volume crap occured\n");
        return -1;
      }

      (*(offset + 1) == 'n') ? (mute = 1) : (mute = 0);

      show_notify_volume(volume, mute);
    }
  }
  else
    return -1;

  return 0;
}

int brightness_stuff(int fd) {
  int nread;

  gint brightness;

  char buf[1024];

  char *bufptr;

  if ( (nread = read(fd, buf, 1024)) > 1) {

    lseek(fd, 0, SEEK_SET);
    buf[nread] = 0x0;

    bufptr = buf;

    for (; !isdigit(*bufptr) && (*bufptr != 0x0); bufptr++);

    if (*bufptr == 0x0) {
      printf("brightness crap occured\n");
      return -1;
    }

    brightness = g_ascii_strtoull(bufptr, NULL, 10);

    show_notify_brightness(brightness);
  }
  else
    return -1;

  return 0;
}

int new_brightness_stuff(int fd) {
  int nread;

  gint brightness;

  char buf[1024];

  char *bufptr;

  nread = read(fd, buf, 1024);

  if ( strncmp("<not supported>\n", buf, nread) == 0 ) {
    perror("new brightness stuff not supported\n");
    return -1;
  }

  if ( nread > 1) {

    lseek(fd, 0, SEEK_SET);
    buf[nread] = 0x0;

    bufptr = strchr(buf, '\n');

    if (bufptr != NULL) {
      for (; !isdigit(*bufptr) && (*bufptr != 0x0); bufptr++);

      if (*bufptr == 0x0) {
        printf("brightness new crap occured\n");
        return -1;
      }

      brightness = g_ascii_strtoull(bufptr, NULL, 10);

      show_notify_brightness(brightness);
    }
  }
  else
    return -1;

  return 0;
}

int fdselect_initialize() {
  fds.inotifyfd   = -1;
  fds.nofd        = 3;
  fds.activefd    = &active_fd[0];
  fds.highestfd   = -1;

  return 0;
}

int open_file(const char* pathname, int flags) {
  int fd;

  if ( (fd = open(pathname, flags)) == -1 ) {
    perror("open\n");
    return -1;
  }

  return fd;
}

int inotify_start() {
  int rc;

  rc = inotify_init_fdselect();
  if(rc < 0)
    return rc;

  build_watch_list();

  return 0;
}

int inotify_init_fdselect() {
  int fd;

  fd = inotify_init();
  if (fd < 0) {
    perror ("inotify_init\n");
    return -1;
  }
  else {
    fds.inotifyfd = fd;
    if(fd > fds.highestfd)
      fds.highestfd = fd;
  }

  return fd;
}

int inotify_add_watch_fdselect(const char* pathname, uint32_t mask, int proc) {
  int fd;
  int wd;

  fd = open_file(pathname, O_RDONLY);
  if(fd < 0)
    return fd;

  wd = inotify_add_watch (fds.inotifyfd, pathname, mask);
  if (wd < 0) {
    perror("inotify_add_watch\n");
    close(fd);
    return -1;
  }
  else {
    fds.activefd[proc].fd = fd;
    fds.activefd[proc].wd = wd;
  }

  return wd;
}

int inotify_get_wd() {
  size_t buf_len = sizeof (struct inotify_event);

  char   buf[ buf_len ];
  int    len;
  struct inotify_event *event;

  len = read (fds.inotifyfd, buf, buf_len);
  if (len < 0) {
    perror ("read\n");
    return -1;
  }
  else if (!len) {
    fprintf(stderr, "buf_len too small ...\n");
    return -1;
  }

  event = (struct inotify_event *) &buf;
  printf ("wd=%d mask=%u cookie=%u len=%u\n",
                                              event->wd, event->mask,
                                              event->cookie, event->len);

  /*
   * According to the manual this should never happen if wd points
   * directly to a file, e.g. /proc/acpi/ibm/volume.
   * */
  if (event->len)
    printf ("name=%s\n", event->name);

  return event->wd;
}

int inotify_reset() {
  int listnum;

  for(listnum = 0; listnum < fds.nofd; listnum++) {
    if( fds.activefd[listnum].fd != -1 ) {
      printf("Closing fd: %d\n", fds.activefd[listnum].fd);
      close(fds.activefd[listnum].fd);
    }
  }

  printf("Closing inotify fd: %d\n", fds.inotifyfd);
  close(fds.inotifyfd);

  return 0;
}

int main(int argc, char *argv[]) {
  int rc;
  int selectfd;
  int listnum;

  int wd;

  signal(SIGINT, signal_handler);
  signal(SIGQUIT, signal_handler);

  rc = fdselect_initialize();
  if(rc < 0 )
    return rc;

  gtk_init(&argc, &argv);


  if(daemon(0, 1) != 0) {
    perror("daemon");
    exit(1);
  }

  while(1) {
    rc = inotify_start();
    if(rc < 0) {
      fprintf(stderr, "can't tolerate wrong initialization, exiting...");
      kill(getpid(), SIGINT);
    }

    build_select_list();
    selectfd = select(fds.highestfd + 1, &fds.triggeredfd, 0, 0, 0);

    if(selectfd < 0) {
      perror("select");
      return -2;
    }
    else if(selectfd == 0) {
      fprintf(stdout, "Time over, connection terminated\n");
      return 0;
    }
    else {
      wd = inotify_get_wd();

      for(listnum = 0; listnum < fds.nofd; listnum++)
        if( fds.activefd[listnum].wd == wd )
          break;

      switch(listnum) {
        case 0:
          volume_stuff(fds.activefd[listnum].fd);
          break;

        case 1:
          brightness_stuff(fds.activefd[listnum].fd);
          break;

        case 2:
          new_brightness_stuff(fds.activefd[listnum].fd);
          break;

        default:
          fprintf(stderr, "\n...esoteric select activity from fd: %d", selectfd);
      }

      inotify_reset();
    }
  }

  return 0;
}

void signal_handler(int signaltype) {
  inotify_reset();
  exit(0);
}

